#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include "../epsilon.h"
#include "../linear_algebra.h"
#include "../linear_programming.h"

enum {
  LP_DEBUG = 0
};

int lp_maximize(struct lp_basis *basis) {
  double eps = epsilon();
  int n_row_ops = 0, e;
#define LP_ARRAY(type, name, dimension)         \
  type *name = basis->name;
#define LP_SCALAR(type, name)                   \
  type name = basis->name;
#include "basis_members.h"
  for (int k = 0; k < m + n; k++) {
    qix[k] = 0.0;
  }
  la_factor(a, p, lu, m);
  while (1) {
    la_forward_solve_transposed(lu, c, yp, m);
    for (int k = n_row_ops - 1; k > -1; k--) {
      int i = row_op[k].i;
      yp[i - 1] -= row_op[k].g * yp[i];
      if (row_op[k].swap) {
        double yp_i = yp[i];
        yp[i] = yp[i - 1];
        yp[i - 1] = yp_i;
      }
    }
    la_backward_solve_transposed(p, lu, yp, y, m);
    if (LP_DEBUG) {
      double *at = calloc(m * m, sizeof *at);
      la_transpose(a, at, m);
      assert(la_relative_residual(at, c, y, m) <= epsilon());
      free(at);
    }
    e = -1;
    double r_e = eps;
    for (int j = m; j < m + n; j++) {
      double r_j = c[j] - la_dot_product(y, a + j * m, m);
      if (r_j > r_e) {
        e = j;
        r_e = r_j;
      }
    }
    la_forward_solve(p, lu, b, x, m);
    for (int k = 0; k < n_row_ops; k++) {
      int i = row_op[k].i;
      if (row_op[k].swap) {
        double x_i = x[i];
        x[i] = x[i - 1];
        x[i - 1] = x_i;
      }
      x[i] -= row_op[k].g * x[i - 1];
    }
    la_backward_solve(lu, x, x, m);
    if (LP_DEBUG) {
      assert(la_relative_residual(a, b, x, m) <= epsilon());
      for (int i = 0; i < m; i++) {
        assert(x[i] >= -eps);
      }
    }
    if (e == -1) {
      for (int j = 0; j < m; j++) {
        qix[q[j]] = x[j];
      }
      break;
    }
    la_forward_solve(p, lu, a + e * m, udx, m);
    for (int k = 0; k < n_row_ops; k++) {
      int i = row_op[k].i;
      if (row_op[k].swap) {
        double udx_i = udx[i];
        udx[i] = udx[i - 1];
        udx[i - 1] = udx_i;
      }
      udx[i] -= row_op[k].g * udx[i - 1];
    }
    la_backward_solve(lu, udx, dx, m);
    if (LP_DEBUG) {
      assert(la_relative_residual(a, a + e * m, dx, m) <= epsilon());
    }
    int l = -1;
    double t_l = INFINITY;
    for (int j = 0; j < m; j++) {
      double dx_j = dx[j];
      if (!(dx_j > eps)) {
        continue;
      }
      double t_j = x[j] / dx_j;
      if (t_j < t_l) {
        l = j;
        t_l = t_j;
      }
    }
    if (l == -1) {
      for (int j = 0; j < m; j++) {
        qix[q[j]] = -dx[j];
      }
      qix[q[e]] = 1.0;
      break;
    }
    for (int i = 0; i < m; i++) {
      double a_i_e = a[e * m + i];
      a[e * m + i] = a[i * m + l];
      for (int j = l + 1; j < m; j++) {
        a[i * m + j - 1] = a[i * m + j];
      }
      a[i * m + m - 1] = a_i_e;
    }
    double c_e = c[e];
    c[e] = c[l];
    for (int j = l + 1; j < m; j++) {
      c[j - 1] = c[j];
    }
    c[m - 1] = c_e;
    int q_e = q[e];
    q[e] = q[l];
    for (int j = l + 1; j < m; j++) {
      q[j - 1] = q[j];
    }
    q[m - 1] = q_e;
    if (n_row_ops <= m * m - (m - (l + 1))) {
      int i;
      for (i = 0; i <= l; i++) {
        for (int j = l + 1; j < m; j++) {
          lu[i * m + j - 1] = lu[i * m + j];
        }
        lu[i * m + m - 1] = udx[i];
      }
      for (; i < m; i++) {
        double h_im1_im1 = lu[(i - 1) * m + i - 1], h_i_im1 = lu[i * m + i];
        if (fabs(h_im1_im1) < fabs(h_i_im1)) {
          double g = h_im1_im1 / h_i_im1;
          lu[(i - 1) * m + i - 1] = h_i_im1;
          for (int j = i + 1; j < m; j++) {
            double h_i_jm1 = lu[i * m + j];
            lu[i * m + j - 1] = lu[(i - 1) * m + j - 1] - g * h_i_jm1;
            lu[(i - 1) * m + j - 1] = h_i_jm1;
          }
          double h_i_mm1 = udx[i];
          lu[i * m + m - 1] = lu[(i - 1) * m + m - 1] - g * h_i_mm1;
          lu[(i - 1) * m + m - 1] = h_i_mm1;
          row_op[n_row_ops++] = (struct lp_row_op){g, i, 1};
        } else {
          double g = h_i_im1 / h_im1_im1;
          for (int j = i + 1; j < m; j++) {
            lu[i * m + j - 1] = lu[i * m + j] - g * lu[(i - 1) * m + j - 1];
          }
          lu[i * m + m - 1] = udx[i] - g * lu[(i - 1) * m + m - 1];
          row_op[n_row_ops++] = (struct lp_row_op){g, i, 0};
        }
      }
    } else {
      n_row_ops = 0;
      la_factor(a, p, lu, m);
    }
  }
  if (LP_DEBUG && e == -1) {
    for (int i = 0; i < m; i++) {
      assert(la_dot_product(a + i * m, x, m) - b[i] <= eps);
    }
    for (int j = 0; j < m; j++) {
      assert(y[j] >= -eps);
    }
    double *at = calloc(m * m, sizeof *at);
    la_transpose(a, at, m);
    for (int j = 0; j < m; j++) {
      assert(c[j] - la_dot_product(at + j * m, y, m) <= eps);
    }
    free(at);
  }
  return e;
}
