#include "linear_programming.h"

#include <float.h>
#include "linear_algebra.h"

static void print_row(const double c[], int n, FILE *stream) {
  for (int j = 0; j < n; j++) {
    if (j) {
      fputs(" + ", stream);
    }
    fprintf(stream, "%.*g x%d", DBL_DIG, c[j], j + 1);
  }
}

static void print_problem(const double a[],
                          const double b[],
                          const double c[],
                          int m,
                          int n,
                          FILE *stream) {
  fputs("Maximize\n obj: ", stream);
  print_row(c, n, stream);
  fputs("\nSubject To\n", stream);
  for (int i = 0; i < m; i++) {
    fprintf(stream, " c%d: ", i + 1);
    print_row(a + i * n, n, stream);
    fprintf(stream, " <= %.*g\n", DBL_DIG, b[i]);
  }
  fputs("End\n", stream);
}

enum {
  M = 20,
  N = 10
};

#define INPUT(name, dimension)                  \
  static double name[dimension];                \
  la_random(&seed, name, dimension)

int main(void) {
  long unsigned int seed = 1;
  INPUT(a, M * N);
  INPUT(b, M);
  INPUT(c, N);
  print_problem(a, b, c, M, N, stdout);
  struct lp_basis basis;
  lp_allocate(M, N, &basis);
  lp_initialize(&basis);
  for (int i = 0; i < M; i++) {
    for (int j = 0; j < N; j++) {
      basis.a[(M + j) * M + i] = a[i * N + j];
    }
  }
  for (int i = 0; i < M; i++) {
    basis.b[i] = b[i];
  }
  for (int j = 0; j < N; j++) {
    basis.c[M + j] = c[j];
  }
  lp_maximize(&basis);
  fprintf(stderr, "%.*g\n", DBL_DIG, la_dot_product(basis.c, basis.x, M));
  fprintf(stderr, "%.*g\n", DBL_DIG, la_dot_product(basis.b, basis.y, M));
  lp_free(&basis);
}
