#include "priority_queue.h"

#include <assert.h>
#include <stdio.h>

static void first_forest(int f[], int n) {
  for (int i = n; i > -1; i--) {
    f[i] = 2 * i + 1;
  }
}

static int next_forest(int f[], int n) {
  int i = 0;
  while (f[i] + 1 >= f[i + 1]) {
    i++;
  }
  f[i]++;
  first_forest(f, i - 1);
  return i + 1 < n;
}

static void print_forest(const int f[], int n) {
  int j = 0;
  for (int i = 0; i < n; i++) {
    while (j++ < f[i]) {
      putchar('(');
    }
    putchar(')');
  }
  putchar('\n');
}

static void test_forest(void) {
  enum {
    N = 4
  };
  int f[N + 1];
  first_forest(f, N);
  do {
    print_forest(f, N);
  } while (next_forest(f, N));
}

static void first_permutation(int p[], int n) {
  for (int i = n; i > -1; i--) {
    p[i] = i;
  }
}

static void swap(int p[], int i, int j) {
  int p_i = p[i];
  p[i] = p[j];
  p[j] = p_i;
}

static void reverse(int p[], int i, int j) {
  while (i < --j) {
    swap(p, i++, j);
  }
}

static int next_permutation(int p[], int n) {
  int j = 1;
  while (p[j] < p[j - 1]) {
    j++;
  }
  int i = 0;
  while (p[j] < p[i]) {
    i++;
  }
  swap(p, i, j);
  reverse(p, 0, j);
  return j < n;
}

static void print_permutation(const int p[], int n) {
  for (int i = n - 1; i > -1; i--) {
    printf("%d%c", p[i], i ? ' ' : '\n');
  }
}

static void test_permutation(void) {
  enum {
    N = 4
  };
  int p[N + 1];
  first_permutation(p, N);
  do {
    print_permutation(p, N);
  } while (next_permutation(p, N));
}

static void insert(double q[], int i, int k, double key) {
  for (; k > i && key < q[k - 1]; k--) {
    q[k] = q[k - 1];
  }
  q[k] = key;
}

int main(void) {
  test_forest();
  test_permutation();
  enum {
    N = 7
  };
  int f[N + 1];
  first_forest(f, N);
  do {
    int p[N + 1];
    first_permutation(p, N);
    do {
      double q[N];
      struct pq_node *root = NULL, node[N];
      int j = 0;
      for (int i = 0; i < N; i++) {
        while (j++ < f[i]) {
          int k = j - i - 1;
          double key = p[k];
          insert(q, i, k, key);
          node[k].key = key;
          root = pq_push(root, &node[k]);
        }
        assert(q[i] == root->key);
        root = pq_pop(root);
        assert((i + 1 >= j - (i + 1)) == !root);
      }
    } while (next_permutation(p, N));
  } while (next_forest(f, N));
}
