#include <stdio.h>
#include "priority_queue.h"

enum {
  SF_LENGTH_MAX = 50,
  SF_WIDTH_MAX = 50,
  SF_FIXED_COST = 10,
  SF_COST_PER_UNIT2 = 1,
  SF_STRAWBERRY = '@'
};

struct sf_input {
  int n_greenhouses_max;
  char field[SF_LENGTH_MAX][SF_WIDTH_MAX];
};

struct sf_interval1 {
  int min, max;
};

struct sf_interval2 {
  struct sf_interval1 i, j;
};

struct sf_greenhouse_list {
  int n_greenhouses;
  struct sf_interval2 *greenhouse;
};

#define SF_EMPTY_GREENHOUSE_LIST ((struct sf_greenhouse_list){0, NULL})

struct sf_point {
  int i, j;
};

struct sf_point_list {
  int n_points;
  struct sf_point point[SF_LENGTH_MAX * SF_WIDTH_MAX];
};

struct sf_label {
  double y;
  struct sf_label *next;
  int mark, n_non_reps;
};

struct sf_equivalence_relation {
  struct sf_label *label[SF_LENGTH_MAX][SF_WIDTH_MAX];
  struct sf_label pool[SF_LENGTH_MAX * SF_WIDTH_MAX + 1];
};

struct sf_node {
  struct pq_node super;
  double y[SF_LENGTH_MAX][SF_WIDTH_MAX];
  double tax;
  double y_sum[SF_LENGTH_MAX + 1][SF_WIDTH_MAX + 1];
  struct sf_greenhouse_list list, solution;
  struct sf_node *previous, *parent, *active;
  struct sf_point p, q;
  int exclude, id;
};

static inline int sf_cost(struct sf_interval2 g) {
  return SF_FIXED_COST + (g.i.max - g.i.min) * (g.j.max - g.j.min) * SF_COST_PER_UNIT2;
}

static inline int sf_int_in_interval1(int p, struct sf_interval1 r) {
  return r.min <= p && p < r.max;
}

static inline int sf_point_in_interval2(struct sf_point p, struct sf_interval2 r) {
  return sf_int_in_interval1(p.i, r.i) && sf_int_in_interval1(p.j, r.j);
}

static inline double sf_sum(double sum[SF_LENGTH_MAX + 1][SF_WIDTH_MAX + 1],
                            struct sf_interval2 r) {
  return sum[r.i.max][r.j.max] - sum[r.i.max][r.j.min] -
    sum[r.i.min][r.j.max] + sum[r.i.min][r.j.min];
}

void sf_activate(double upper_bound, struct sf_node *node);
void sf_append(struct sf_interval2 g, struct sf_greenhouse_list *list);
int sf_bounding_box_from_input(const struct sf_input *input,
                               struct sf_interval2 *box);
void sf_choose_representatives(struct sf_interval2 box,
                               const struct sf_equivalence_relation *relation,
                               const struct sf_point_list *s_list,
                               struct sf_point_list *rep_list);
void sf_fast_simplex(const struct sf_input *input,
                     struct sf_interval2 box,
                     double upper_bound,
                     struct sf_node *node);
struct sf_greenhouse_list sf_copy(struct sf_greenhouse_list list);
struct sf_greenhouse_list sf_greenhouse_list_from_input(const struct sf_input *input);
void sf_label_equivalence_classes(struct sf_interval2 box,
                                  struct sf_greenhouse_list solution,
                                  struct sf_equivalence_relation *relation);
struct sf_greenhouse_list sf_partition(struct sf_node *node,
                                       double upper_bound,
                                       struct sf_greenhouse_list list);
void sf_primal_dual(const struct sf_input *input,
                    struct sf_interval2 box,
                    double upper_bound,
                    struct sf_node *node);
int sf_read_input(FILE *stream, struct sf_input *input);
int sf_select_branch(struct sf_interval2 box,
                     struct sf_greenhouse_list solution,
                     struct sf_node *node);
void sf_simplex(const struct sf_input *input,
                struct sf_interval2 box,
                double upper_bound,
                struct sf_node *node);
void sf_solve(const struct sf_input *input, struct sf_greenhouse_list *solution);
void sf_strawberry_list_from_input(const struct sf_input *input,
                                   struct sf_point_list *s_list);
void sf_sums_from_prices(struct sf_interval2 box,
                         double y[SF_LENGTH_MAX][SF_WIDTH_MAX],
                         double sum[SF_LENGTH_MAX + 1][SF_WIDTH_MAX + 1]);
void sf_verify_solution(const struct sf_input *input,
                        struct sf_greenhouse_list solution);
void sf_write_solution(const struct sf_input *input,
                       struct sf_greenhouse_list solution,
                       FILE *stream);
