#include "../strawberry_fields.h"

static int merge(const int *restrict a, int m, const int *restrict b, int n,
                 int *restrict c) {
  int i = 0, j = 0, k = 0;
  while (1) {
    if (a[i] <= b[j]) {
      c[k++] = a[i++];
      if (i >= m) {
        do {
          c[k++] = b[j++];
        } while (j < n);
        break;
      }
    } else {
      c[k++] = b[j++];
      if (j >= n) {
        do {
          c[k++] = a[i++];
        } while (i < m);
        break;
      }
    }
  }
  return k;
}

static int unique(const int *restrict a, int m, int *restrict b) {
  *b = *a;
  int j = 1;
  for (int i = 1; i < m; i++) {
    if (a[i] != b[j - 1]) {
      b[j++] = a[i];
    }
  }
  return j;
}

static int union_update(const int *restrict a, int m, int *restrict b, int n) {
  if (n) {
    int c[m + n];
    return unique(c, merge(a, m, b, n, c), b);
  } else {
    for (int i = 0; i < m; i++) {
      b[i] = a[i];
    }
    return m;
  }
}

struct sf_greenhouse_list sf_greenhouse_list_from_input(const struct sf_input *input) {
  struct sf_greenhouse_list list = SF_EMPTY_GREENHOUSE_LIST;
  int begin[SF_LENGTH_MAX + 1], m = 0, column[SF_LENGTH_MAX * SF_WIDTH_MAX],
    cum_sum[SF_LENGTH_MAX][SF_WIDTH_MAX + 2];
  for (int i = 0; i < SF_LENGTH_MAX; i++) {
    begin[i] = m;
    int sum = 0;
    for (int j = 0; j < SF_WIDTH_MAX; j++) {
      cum_sum[i][j] = sum;
      if (input->field[i][j] == SF_STRAWBERRY) {
        column[m++] = j;
        sum++;
      }
    }
    for (int j = 0; j < 2; j++) {
      cum_sum[i][SF_WIDTH_MAX + j] = sum + j;
    }
  }
  begin[SF_LENGTH_MAX] = m;
  for (int i_min = 0; i_min < SF_LENGTH_MAX; i_min++) {
    const int *end = begin + 1;
    if (begin[i_min] >= end[i_min]) {
      continue;
    }
    int active[SF_WIDTH_MAX + 1], n_active = 0;
    for (int i_max = i_min; i_max < SF_LENGTH_MAX; i_max++) {
      if (begin[i_max] >= end[i_max]) {
        continue;
      }
      n_active = union_update(column + begin[i_max], end[i_max] - begin[i_max],
                              active, n_active);
      active[n_active] = SF_WIDTH_MAX;
      int k_max_min = 0;
      for (int k_min = 0; k_min < n_active; k_min++) {
        int j_min = active[k_min];
        while (cum_sum[i_min][active[k_max_min] + 1] <= cum_sum[i_min][j_min]) {
          k_max_min++;
        }
        while (cum_sum[i_max][active[k_max_min] + 1] <= cum_sum[i_max][j_min]) {
          k_max_min++;
        }
        for (int k_max = k_max_min; k_max < n_active; k_max++) {
          int j_max = active[k_max];
          sf_append((struct sf_interval2){{i_min, i_max + 1}, {j_min, j_max + 1}}, &list);
        }
      }
    }
  }
  return list;
}
