#include <math.h>
#include "../epsilon.h"
#include "../strawberry_fields.h"

void sf_primal_dual(const struct sf_input *input,
                    struct sf_interval2 box,
                    double upper_bound,
                    struct sf_node *node) {
  double eps = epsilon();
  upper_bound *= 1.0 + eps;
  double dy[SF_LENGTH_MAX][SF_WIDTH_MAX];
  for (int i = box.i.min; i < box.i.max; i++) {
    for (int j = box.j.min; j < box.j.max; j++) {
      dy[i][j] = input->field[i][j] == SF_STRAWBERRY;
    }
  }
  for (int k = 0; k < node->solution.n_greenhouses; k++) {
    struct sf_interval2 g = node->solution.greenhouse[k];
    for (int i = g.i.min; i < g.i.max; i++) {
      for (int j = g.j.min; j < g.j.max; j++) {
        dy[i][j] = 0.0;
      }
    }
  }
  while (1) {
    double dy_sum[SF_LENGTH_MAX + 1][SF_WIDTH_MAX + 1];
    sf_sums_from_prices(box, node->y, node->y_sum);
    double box_y = sf_sum(node->y_sum, box);
    sf_sums_from_prices(box, dy, dy_sum);
    double box_dy = sf_sum(dy_sum, box);
    if (box_dy <= 0.0) {
      node->super.key = ceil((1.0 - eps) * box_y);
      break;
    }
    int unbounded = 1;
    double gap = upper_bound - box_y, dt_first = gap / box_dy;
    struct sf_interval2 g_first;
    for (int k = 0; k < node->list.n_greenhouses; k++) {
      struct sf_interval2 g = node->list.greenhouse[k];
      double g_dy = sf_sum(dy_sum, g);
      if (g_dy <= 0.0) {
        continue;
      }
      double g_reduced_cost = sf_cost(g) - sf_sum(node->y_sum, g);
      if (g_reduced_cost >= gap) {
        continue;
      }
      double dt = g_reduced_cost / g_dy;
      if (dt < dt_first) {
        unbounded = 0;
        dt_first = dt;
        g_first = g;
      }
    }
    if (unbounded) {
      node->super.key = INFINITY;
      break;
    }
    sf_append(g_first, &node->solution);
    for (int i = box.i.min; i < box.i.max; i++) {
      for (int j = box.j.min; j < box.j.max; j++) {
        node->y[i][j] += dt_first * dy[i][j];
      }
    }
    for (int i = g_first.i.min; i < g_first.i.max; i++) {
      for (int j = g_first.j.min; j < g_first.j.max; j++) {
        dy[i][j] = 0.0;
      }
    }
  }
}
