#include <math.h>
#include "../epsilon.h"
#include "../linear_algebra.h"
#include "../linear_programming.h"
#include "../strawberry_fields.h"

enum {
  SF_TAX = 1
};

void sf_simplex(const struct sf_input *input,
                struct sf_interval2 box,
                double upper_bound,
                struct sf_node *node) {
  double eps = epsilon();
  struct sf_point_list s_list;
  sf_strawberry_list_from_input(input, &s_list);
  while (1) {
    struct sf_equivalence_relation relation;
    sf_label_equivalence_classes(box, node->solution, &relation);
    struct sf_point_list rep_list;
    sf_choose_representatives(box, &relation, &s_list, &rep_list);
    int m = node->solution.n_greenhouses, n = rep_list.n_points;
    struct lp_basis basis;
    lp_allocate(m, n + SF_TAX, &basis);
    lp_initialize(&basis);
    for (int j = 0; j < n; j++) {
      for (int i = 0; i < m; i++) {
        basis.a[(m + j) * m + i] = sf_point_in_interval2(rep_list.point[j], node->solution.greenhouse[i]);
      }
    }
    for (int i = 0; i < m; i++) {
      basis.b[i] = sf_cost(node->solution.greenhouse[i]);
    }
    for (int j = 0; j < n; j++) {
      basis.c[m + j] = 1.0;
    }
    if (SF_TAX) {
      for (int i = 0; i < m; i++) {
        basis.a[(m + n) * m + i] = -1.0;
      }
      basis.c[m + n] = -input->n_greenhouses_max;
    }
    int e = lp_maximize(&basis);
    for (int k = 0; k < rep_list.n_points; k++) {
      struct sf_point p = rep_list.point[k];
      struct sf_label *label = relation.label[p.i][p.j];
      label->y = basis.qix[m + k] / (label->n_non_reps + 1);
    }
    node->tax = SF_TAX ? basis.qix[m + n] : 0.0;
    for (int k = 0; k < s_list.n_points; k++) {
      struct sf_point p = s_list.point[k];
      node->y[p.i][p.j] = relation.label[p.i][p.j]->y;
    }
    sf_sums_from_prices(box, node->y, node->y_sum);
    struct sf_interval2 g_mvc;
    double g_mvc_badness = eps;
    if (e != -1) {
      for (int k = 0; k < node->list.n_greenhouses; k++) {
        struct sf_interval2 g = node->list.greenhouse[k];
        double g_badness = sf_sum(node->y_sum, g) - node->tax;
        if (g_badness > g_mvc_badness) {
          g_mvc = g;
          g_mvc_badness = g_badness;
        }
      }
      if (!(g_mvc_badness > eps)) {
        node->super.key = INFINITY;
        lp_free(&basis);
        break;
      }
    } else {
      for (int k = 0; k < node->list.n_greenhouses; k++) {
        struct sf_interval2 g = node->list.greenhouse[k];
        double g_badness = sf_sum(node->y_sum, g) - node->tax - sf_cost(g);
        if (g_badness > g_mvc_badness) {
          g_mvc = g;
          g_mvc_badness = g_badness;
        }
      }
      if (!(g_mvc_badness > eps)) {
        node->super.key = ceil((1.0 - eps) * la_dot_product(basis.c, basis.x, m));
        int n = 0;
        for (int k = 0; k < m; k++) {
          if (basis.y[k] > eps) {
            node->solution.greenhouse[n++] = node->solution.greenhouse[k];
          }
        }
        node->solution.n_greenhouses = n;
        lp_free(&basis);
        break;
      }
    }
    sf_append(g_mvc, &node->solution);
  }
}
