#include "strawberry_fields.h"

#include <assert.h>
#include <stdlib.h>

static void test_sums(void) {
  double y[SF_LENGTH_MAX][SF_WIDTH_MAX];
  double y_sum[SF_LENGTH_MAX + 1][SF_WIDTH_MAX + 1];
  for (int i = 0; i < SF_LENGTH_MAX; i++) {
    for (int j = 0; j < SF_WIDTH_MAX; j++) {
      y[i][j] = rand();
    }
  }
  sf_sums_from_prices((struct sf_interval2){{0, SF_LENGTH_MAX}, {0, SF_WIDTH_MAX}}, y, y_sum);
  for (long int k = 0; k < 1000000; k++) {
    int i_min = rand() % SF_LENGTH_MAX;
    int i_max = rand() % SF_LENGTH_MAX + 1;
    if (i_min > i_max) {
      int t = i_min;
      i_min = i_max;
      i_max = t;
    }
    int j_min = rand() % SF_WIDTH_MAX;
    int j_max = rand() % SF_WIDTH_MAX + 1;
    if (j_min > j_max) {
      int t = j_min;
      j_min = j_max;
      j_max = t;
    }
    double sum = 0.0;
    for (int i = i_min; i < i_max; i++) {
      for (int j = j_min; j < j_max; j++) {
        sum += y[i][j];
      }
    }
    assert(sum == sf_sum(y_sum, (struct sf_interval2){{i_min, i_max}, {j_min, j_max}}));
  }
}

static void write_set(const struct sf_point_list *s_list,
                      struct sf_interval2 r) {
  for (int k = 0; k < s_list->n_points; k++) {
    putchar(sf_point_in_interval2(s_list->point[k], r) ? '+' : '-');
  }
  putchar('\n');
}

static void write_sets(const struct sf_point_list *s_list) {
  for (int i_min = 0; i_min < SF_LENGTH_MAX; i_min++) {
    for (int i_max = i_min; i_max < SF_LENGTH_MAX; i_max++) {
      for (int j_min = 0; j_min < SF_WIDTH_MAX; j_min++) {
        for (int j_max = j_min; j_max < SF_WIDTH_MAX; j_max++) {
          write_set(s_list, (struct sf_interval2){{i_min, i_max}, {j_min, j_max}});
        }
      }
    }
  }
}

int main(void) {
  test_sums();
  struct sf_input input;
  struct sf_greenhouse_list solution;
  solution.n_greenhouses = 0;
  while (sf_read_input(stdin, &input)) {
    struct sf_greenhouse_list g_list = sf_greenhouse_list_from_input(&input);
    fprintf(stderr, "%d\n", g_list.n_greenhouses);
    switch (2) {
    case 0:
      sf_write_solution(&input, solution, stdout);
      break;
    case 1:
      {
        struct sf_point_list s_list;
        sf_strawberry_list_from_input(&input, &s_list);
        if (1) {
          write_sets(&s_list);
        } else {
          for (int k = 0; k < g_list.n_greenhouses; k++) {
            write_set(&s_list, g_list.greenhouse[k]);
          }
        }
      }
      break;
    }
    free(g_list.greenhouse);
  }
}
